import unittest
import sys
import os

# Add the project root to sys.path so we can import pycp
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from pycp.bit_utils import (
    set_bit, clear_bit, toggle_bit, check_bit, count_set_bits, lowest_set_bit, is_power_of_two
)

class TestBitUtils(unittest.TestCase):
    def test_set_bit(self):
        self.assertEqual(set_bit(0, 0), 1)
        self.assertEqual(set_bit(0, 2), 4)
        self.assertEqual(set_bit(4, 0), 5)
        self.assertEqual(set_bit(4, 2), 4)

    def test_clear_bit(self):
        self.assertEqual(clear_bit(7, 0), 6)
        self.assertEqual(clear_bit(7, 1), 5)
        self.assertEqual(clear_bit(4, 2), 0)
        self.assertEqual(clear_bit(4, 0), 4)

    def test_toggle_bit(self):
        self.assertEqual(toggle_bit(0, 0), 1)
        self.assertEqual(toggle_bit(1, 0), 0)
        self.assertEqual(toggle_bit(5, 1), 7)
        self.assertEqual(toggle_bit(7, 1), 5)

    def test_check_bit(self):
        self.assertTrue(check_bit(5, 0))  # 101 -> 0th bit is 1
        self.assertFalse(check_bit(5, 1)) # 101 -> 1st bit is 0
        self.assertTrue(check_bit(5, 2))  # 101 -> 2nd bit is 1

    def test_count_set_bits(self):
        self.assertEqual(count_set_bits(0), 0)
        self.assertEqual(count_set_bits(1), 1)
        self.assertEqual(count_set_bits(7), 3)
        self.assertEqual(count_set_bits(12345), bin(12345).count('1'))

    def test_lowest_set_bit(self):
        self.assertEqual(lowest_set_bit(12), 4) # 1100 -> 100 which is 4
        self.assertEqual(lowest_set_bit(6), 2)  # 110 -> 10 which is 2
        self.assertEqual(lowest_set_bit(1), 1)
        self.assertEqual(lowest_set_bit(0), 0)

    def test_is_power_of_two(self):
        self.assertTrue(is_power_of_two(1))
        self.assertTrue(is_power_of_two(2))
        self.assertTrue(is_power_of_two(4))
        self.assertTrue(is_power_of_two(1024))
        self.assertFalse(is_power_of_two(0))
        self.assertFalse(is_power_of_two(3))
        self.assertFalse(is_power_of_two(6))

if __name__ == '__main__':
    unittest.main()



import unittest
import sys
import os

# Add the project root to sys.path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from pycp.arrays import (
    get_prefix_sum, query_range_sum, DifferenceArray,
    sliding_window_sum, sliding_window_max, max_subarray_sum, two_sum_sorted
)

class TestArrayUtils(unittest.TestCase):
    def test_prefix_sum(self):
        arr = [1, 2, 3, 4, 5]
        ps = get_prefix_sum(arr)
        self.assertEqual(ps, [0, 1, 3, 6, 10, 15])
        self.assertEqual(query_range_sum(ps, 0, 2), 6) # 1+2+3
        self.assertEqual(query_range_sum(ps, 1, 3), 9) # 2+3+4
        self.assertEqual(query_range_sum(ps, 2, 2), 3) # 3

    def test_difference_array(self):
        arr = [1, 2, 3, 4, 5]
        da = DifferenceArray(arr)
        da.update(0, 2, 2) # [3, 4, 5, 4, 5]
        da.update(1, 4, 1) # [3, 5, 6, 5, 6]
        self.assertEqual(da.get_array(), [3, 5, 6, 5, 6])

    def test_sliding_window_sum(self):
        arr = [1, 3, 2, 6, -1, 4, 1, 8, 2]
        self.assertEqual(sliding_window_sum(arr, 5), [11, 14, 12, 18, 14])
        self.assertEqual(sliding_window_sum([1, 2], 3), [])
        
    def test_sliding_window_max(self):
        arr = [1, 3, -1, -3, 5, 3, 6, 7]
        self.assertEqual(sliding_window_max(arr, 3), [3, 3, 5, 5, 6, 7])
        self.assertEqual(sliding_window_max([1], 1), [1])
        
    def test_kadane(self):
        self.assertEqual(max_subarray_sum([-2, 1, -3, 4, -1, 2, 1, -5, 4]), 6)
        self.assertEqual(max_subarray_sum([-1, -2, -3]), -1)
        self.assertEqual(max_subarray_sum([1, 2, 3]), 6)
        self.assertEqual(max_subarray_sum([]), 0)

    def test_two_sum_sorted(self):
        arr = [2, 7, 11, 15]
        self.assertEqual(two_sum_sorted(arr, 9), (0, 1))
        self.assertEqual(two_sum_sorted(arr, 20), None)
        self.assertEqual(two_sum_sorted([2, 3, 4], 6), (0, 2))

if __name__ == '__main__':
    unittest.main()



import unittest
import sys
import os

# Add the project root to sys.path
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

from pycp.strings import (
    compute_pi, kmp_search, z_function, RollingHash, is_palindrome, manacher
)

class TestStringUtils(unittest.TestCase):
    def test_kmp(self):
        text = "ABABDABACDABABCABAB"
        pattern = "ABABCABAB"
        # Matches at index 10
        self.assertEqual(kmp_search(text, pattern), [10])
        
        text2 = "AAAAA"
        pattern2 = "AA"
        # Matches at 0, 1, 2, 3
        self.assertEqual(kmp_search(text2, pattern2), [0, 1, 2, 3])
        
        self.assertEqual(kmp_search("ABC", "DEF"), [])
        self.assertEqual(compute_pi("ABABAC"), [0, 0, 1, 2, 3, 0])

    def test_z_function(self):
        s = "aabcaabxaaaz"
        # z array manually: 
        # a a b c a a b x a a a z
        # - 1 0 0 3 1 0 0 2 2 1 0 
        # Note: z[0] is usually 0 in standard implementations or undefined, our implementation might leave it as 0 or handle it.
        # Let's check a simpler one.
        s2 = "aaaaa"
        # z: 0 4 3 2 1
        self.assertEqual(z_function(s2), [0, 4, 3, 2, 1])
        
        s3 = "abacaba"
        # z: 0 0 1 0 3 0 1
        self.assertEqual(z_function(s3), [0, 0, 1, 0, 3, 0, 1])

    def test_rolling_hash(self):
        s = "abcdeabcde"
        rh = RollingHash(s)
        # abcde at 0 and abcde at 5 should match
        self.assertEqual(rh.get_hash(0, 4), rh.get_hash(5, 9))
        self.assertNotEqual(rh.get_hash(0, 4), rh.get_hash(0, 3))
        
        # Test collision resistance (basic)
        s2 = "ab"
        s3 = "ba"
        rh2 = RollingHash(s2)
        rh3 = RollingHash(s3)
        self.assertNotEqual(rh2.get_hash(0, 1), rh3.get_hash(0, 1))

    def test_is_palindrome(self):
        self.assertTrue(is_palindrome("aba"))
        self.assertTrue(is_palindrome("abba"))
        self.assertFalse(is_palindrome("abc"))
        self.assertTrue(is_palindrome("a"))
        self.assertTrue(is_palindrome(""))

    def test_manacher(self):
        s = "aba"
        # T = ^#a#b#a#$
        # P = 010131010 (approx, depends on implementation boundary)
        # Our implementation uses ^ and $ sentinels.
        # T: ^ # a # b # a # $
        # i: 0 1 2 3 4 5 6 7 8
        # P: 0 1 0 1 3 1 0 1 0
        p = manacher(s)
        self.assertEqual(p[4], 3) # Center at 'b' (#b#), radius 3 -> #a#b#a#
        
        s2 = "abba"
        # T: ^ # a # b # b # a # $
        # i: 0 1 2 3 4 5 6 7 8 9 10
        # P: 0 1 0 1 0 4 0 1 0 1 0 
        p2 = manacher(s2)
        self.assertEqual(p2[5], 4) # Center at '#' between b and b, radius 4 -> #a#b#b#a#

if __name__ == '__main__':
    unittest.main()
